/*
 * Copyright 2006-2020 The MZmine Development Team
 * 
 * This file is part of MZmine.
 * 
 * MZmine is free software; you can redistribute it and/or modify it under the terms of the GNU
 * General Public License as published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * MZmine is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with MZmine; if not,
 * write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
 * USA
 */

package io.github.mzmine.modules.visualization.chromatogram;

import io.github.mzmine.datamodel.features.Feature;
import java.text.NumberFormat;
import java.util.Map;

import org.jfree.chart.labels.XYToolTipGenerator;
import org.jfree.data.xy.XYDataset;

import io.github.mzmine.datamodel.FeatureInformation;
import io.github.mzmine.main.MZmineCore;

/**
 * Tooltip generator for TIC visualizer
 */
public class TICToolTipGenerator implements XYToolTipGenerator {

  private final NumberFormat rtFormat = MZmineCore.getConfiguration().getRTFormat();
  private final NumberFormat mzFormat = MZmineCore.getConfiguration().getMZFormat();
  private final NumberFormat intensityFormat = MZmineCore.getConfiguration().getIntensityFormat();

  @Override
  public String generateToolTip(final XYDataset dataSet, final int series, final int item) {

    final String toolTip;

    final double rtValue = dataSet.getXValue(series, item);
    final double intValue = dataSet.getYValue(series, item);

    if (dataSet instanceof TICDataSet) {

      final TICDataSet ticDataSet = (TICDataSet) dataSet;

      toolTip = "Scan #" + ticDataSet.getScan(item).getScanNumber() + "\nRetention time: "
          + rtFormat.format(rtValue) + "\nBase peak m/z: "
          + mzFormat.format(ticDataSet.getZValue(series, item)) + "\nIntensity: "
          + intensityFormat.format(intValue);

    } else if (dataSet instanceof FeatureDataSet) {

      final FeatureDataSet featureDataSet = (FeatureDataSet) dataSet;
      final Feature feature = featureDataSet.getFeature();
      FeatureInformation featureInfo = null;
      if (feature != null) {
        featureInfo = feature.getFeatureInformation();
      }

      final String label = featureDataSet.getName();
      String text = label == null || label.length() == 0 ? "" : label + '\n';
      text += "Retention time: " + rtFormat.format(rtValue) + "\nm/z: "
          + mzFormat.format(featureDataSet.getMZ(item)) + "\nIntensity: "
          + intensityFormat.format(intValue);

      NumberFormat numberFormat = NumberFormat.getInstance();

      if (featureInfo != null)
        for (Map.Entry<String, String> e : featureInfo.getAllProperties().entrySet()) {
          try {
            double value = Double.parseDouble(e.getValue());
            text += "\n" + e.getKey() + ": " + numberFormat.format(value);
          } catch (NullPointerException | NumberFormatException exception) {
            continue;
          }
        }

      toolTip = text;

    } else {

      toolTip = "Retention time: " + rtFormat.format(rtValue) + "\nIntensity: "
          + intensityFormat.format(intValue);
    }

    return toolTip;
  }
}
